"""

"""

import sys
import types


def init_args(cls):
    """ Return the __init__ args (minus 'self') for @cls

    Args:
        cls: class, instance or callable
    Returns:
        list of str, the arguments minus 'self'
    """
    # This looks insanely goofy, but seems to literally be the
    # only thing that actually works.  Your obvious ways to
    # accomplish this task do not apply here.
    try:
        # Assume it's a factory function, static method, or other callable
        args = cls.__code__.co_varnames
    except AttributeError:
        # assume it's a class
        args = cls.__init__.__code__.co_varnames

    # Note:  There is a special place in hell for people who don't
    #        call the first method argument 'self'.
    if args[0] == 'self':
        args = args[1:]

    return args




"""

"""

from .util.marshal import *
from .util.type import *


__all__ = [
    'ExtraKeysError',
    'InitArgsError',
    'marshal_json',
    'type_assert',
    'type_assert_dict',
    'type_assert_iter',
    'unmarshal_json',
]

JSON_TYPES = (
    bool,
    dict,
    float,
    int,
    list,
    str,
    type(None),
)


def marshal_json(
    obj,
    types=JSON_TYPES,
    fields=None,
):
    """ Recursively marshal a Python object to a JSON-compatible dict
        that can be passed to json.{dump,dumps}, a web client,
        or a web server, etc...

    Args:
        obj:    object, It's members can be nested Python
                objects which will be converted to dictionaries
        types:  tuple-of-types, The JSON primitive types, typically
                you would not change this
        fields: None-list-of-str, Explicitly marshal only these fields
    Returns:
        dict
    """
    return marshal_dict(
        obj,
        types,
        fields=fields,
    )


def unmarshal_json(
    obj,
    cls,
    allow_extra_keys=True,
    ctor=None,
):
    """ Unmarshal @obj into @cls

    Args:
        obj:              dict, A JSON object
        cls:              type, The class to unmarshal into
        allow_extra_keys: bool, False to raise an exception when extra
                          keys are present, True to ignore
        ctor:             None-or-static-method: Use this method as the
                          constructor instead of __init__
    Returns:
        instance of @cls
    Raises:
        ExtraKeysError: If allow_extra_keys == False, and extra keys
                        are present in @obj and not in @cls.__init__
        ValueError:     If @cls.__init__ does not contain a self argument
    """
    return unmarshal_dict(
        obj,
        cls,
        allow_extra_keys,
        ctor=ctor,
    )



"""

"""

from .init_args import init_args
from .key_swap import key_swap


__all__ = [
    'ExtraKeysError',
    '_get_dict',
    'InitArgsError',
    'marshal_dict',
    'unmarshal_dict',
    'marshal_list',
    'unmarshal_list',
]


def _get_dict(obj):
    """ Hack to work around the lack of __dict__ when __slots__ is used

    """
    if isinstance(obj, dict):
        return False, obj
    has_slots = hasattr(obj, '__slots__')
    if has_slots:
        d = {k:getattr(obj, k) for k in obj.__slots__}
    else:
        d = obj.__dict__
    return has_slots, d


class ExtraKeysError(Exception):
    """ Raised when extra object keys are present

        This exception can be marshalled into JSON for sending
        to clients.  However, it cannot be unmarshalled back into
        an ExtraKeysError
    """
    def __init__(
        self,
        cls,
        diff,
    ):
        """ Note that type_assert can't be used because it would
            create a circular dependency.

        Args:
            cls,  type, The type that was attempted to unmarshal into
            diff: dict, The extra arguments that were passed to @cls
        """
        msg = "\n".join([
            "", # Newline to make the output cleaner
            "ctor: {}".format(cls),
            "extras: {}".format(diff)
        ])
        Exception.__init__(self, msg)
        self.type = str(
            type(self),
        )
        self.cls = str(cls)
        self.diff = str(diff)
        self.type = self.__class__.__name__


class InitArgsError(Exception):
    """ Raised when unmarshalling a class raises an Exception

        This exception can be marshalled into JSON for sending
        to clients.  However, it cannot be unmarshalled back into
        an InitArgsError
    """
    def __init__(
        self,
        cls,
        cls_args,
        kwargs,
        ex,
    ):
        """ Note that type_assert can't be used because it would
            create a circular dependency.

        Args:
            cls,      type-or-static-method, The type or constructor
                      that was attempted to unmarshal into
            cls_args: list, The arguments of @cls
            kwargs:   dict, The arguments that were passed to @cls
            ex:       Exception, The exception that was raised
        """
        msg = "\n".join([
            "", # Newline to make the output cleaner
            "module: {}".format(cls.__module__),
            "ctor: {}".format(cls),
            "ctor_args: {}".format(cls_args),
            "args (after removing args not in ctor_args): {}".format(kwargs),
            "only in ctor_args".format(
                [x for x in cls_args if x not in kwargs]
            ),
            "exception: {}".format(ex),
        ])
        Exception.__init__(self, msg)
        self.type = str(
            type(self),
        )
        self.cls = str(cls)
        self.cls_args = str(cls_args)
        self.kwargs = str(kwargs)
        self.ex = str(ex)
        self.type = self.__class__.__name__

def _marshal_value(
    k,
    v,
    types,
    method,
    **m_kwargs
):
    if isinstance(v, dict):
        return marshal_dict(v, types)
    elif isinstance(v, list):
        return _marshal_list(v, types)
    elif isinstance(v, types):
        return v
    elif method:
        return getattr(v, method)(**m_kwargs)
    else:
        return marshal_dict(v, types)

def _marshal_list(
    _list,
    types,
):
    return [
        x if isinstance(x, (bool, float, int, str, type(None)))
        else marshal_dict(x, types)
        for x in _list
    ]

def marshal_dict(
    obj,
    types,
    method=None,
    fields=None,
    **m_kwargs
):
    """ Recursively marshal a Python object to a dict
        that can be passed to json.{dump,dumps}, a web client,
        or a web server, document database, etc...

    Args:
        obj:      object, It's members can be nested Python
                  objects which will be converted to dictionaries
        types:    tuple-of-types, The primitive types that can be
                  serialized
        method:   None-or-str, None to use 'marshal_dict' recursively,
                  or a str that corresponds to the name of a class method
                  to use.  Any nested types that are not an instance of
                  @types must have this method defined.
        fields:   None-list-of-str, Explicitly marshal only these fields
        m_kwargs: Keyword arguments to pass to @method
    Returns:
        dict
    """
    result = {}
    has_slots, d = _get_dict(obj)

    if fields:
        for field in fields:
            assert field in d
        return {
            k: _marshal_value(k, v, types, method, **m_kwargs)
            for k, v in d.items()
            if k in fields
        }

    excl = getattr(obj, '_marshal_exclude', [])

    if (
        has_slots
        or
        getattr(obj, '_marshal_only_init_args', False)
    ):
        args = init_args(obj)
        excl.extend([x for x in d if x not in args])

    if getattr(obj, '_marshal_exclude_none', False):
        excl.extend(k for k, v in d.items() if v is None)
    else:
        none_keys = getattr(obj, '_marshal_exclude_none_keys', [])
        if none_keys:
            excl.extend(x for x in none_keys if d.get(x) is None)

    return {
        k: _marshal_value(k, v, types, method, **m_kwargs)
        for k, v in d.items()
        if k not in excl
    }

def unmarshal_dict(
    obj,
    cls,
    allow_extra_keys=True,
    ctor=None,
):
    """ Unmarshal @obj into @cls

    Args:
        obj:              dict, The dict to unmarshal into @cls
        cls:              type, The class to unmarshal into
        allow_extra_keys: bool, False to raise an exception when extra
                          keys are present, True to ignore
        ctor:             None-or-method:  Use this as a constructor instead
                          of __init__
    Returns:
        instance of @cls
    Raises:
        ExtraKeysError: If allow_extra_keys == False, and extra keys
                        are present in @obj and not in @cls.__init__
        ValueError:     If @cls.__init__ does not contain a self argument
    """
    if not ctor:
        ctor = cls
    args = init_args(ctor)
    obj = key_swap(obj, cls, False)
    kwargs = {k: v for k, v in obj.items() if k in args}

    # If either is set to False, do not allow extra keys
    # to be present in obj but not in cls.__init__
    allow_extra_keys = (
        getattr(cls, '_unmarshal_allow_extra_keys', True)
        and
        allow_extra_keys
    )

    if not allow_extra_keys and len(obj) > len(kwargs):
        diff = {k: v for k, v in obj.items() if k not in args}
        raise ExtraKeysError(cls, diff)

    try:
        return ctor(**kwargs)
    except ExtraKeysError as ex:
        raise ex
    except Exception as ex:
        raise InitArgsError(
            cls,
            args,
            kwargs,
            ex,
        )


def marshal_list(
    obj,
    cls,
    types,
    fields=None
):
    """ Marshal @obj into a list
    Args:
        obj:   object, the object to marshal into a list
        cls:   function-or-method, Use this as a constructor instead
               of __init__
        types: iterable, a collection of valid types
    Returns:
        list
    Raises:
        ValueError: If the class fields contain invalid types
    """
    if isinstance(cls, type):
        cls = cls.__init__
    if not fields:
        fields = cls.__code__.co_varnames
        if fields[0] == 'self':
            fields = fields[1:]
    result = [
        getattr(obj, arg)
        for arg in fields
    ]
    invalid = [
        x for x in result
        if type(x) not in types
    ]
    if invalid:
        raise ValueError("Invalid fields: {}".format(invalid))
    return result

def unmarshal_list(
    _list,
    cls,
):
    """ Unmarshal @_list into @cls

    Args:
        _list: list-or-tuple, The list to unmarshal into @cls
        cls:   type-or-function, The class to unmarshal into
    Returns:
        List of instances of @cls
    Raises:
        InitArgsError: If instantiation fails
    """
    try:
        return cls(*_list)
    except Exception as ex:
        raise InitArgsError(
            cls,
            _list,
            {},
            ex,
        )



"""
    Various utility functions to supplement the functionality of
    the type assert functions
"""


__all__ = [
    'check_dups',
]


def check_dups(
    iterable,
    debug_limit=1000,
):
    """ Checks an iterable for duplicates

        Note that it does not make sense to call this on a set()

        If calling on a collection without a .count method,
        set @debug_limit to -1.

        For custom equality comparisons, create a custom
        __eq__ and __hash__ method in the class contained in @iterable

    Args:
        iterable:    An iterable to check for duplicates
        debug_limit: Don't generate a list of duplicates for the
                     Exception message if the length of @iterable
                     is >= this value.  This exists because generating
                     the list does not scale efficiently.
    Raises:
        ValueError: If @iterable contains duplicates
    """
    if not iterable:
        return
    unique = set(iterable)
    if len(unique) != len(iterable):
        if len(iterable) < debug_limit:
            dups = [x for x in unique if iterable.count(x) > 1]
            msg = "Duplicate values: {}".format(dups)
        else:
            msg = (
                "Duplicate values, not generating list because "
                "len(iterable)={} is greater than debug_limit={}"
            ).format(
                len(iterable),
                debug_limit,
            )
        raise ValueError(msg)




"""

"""

from .util.marshal import *
from .util.type import *


__all__ = [
    'InitArgsError',
    'marshal_csv',
    'type_assert',
    'type_assert_iter',
    'unmarshal_csv',
]

CSV_TYPES = (
    float,
    int,
    str,
)

def marshal_csv(
    iterable,
    cls,
    types=CSV_TYPES,
    fields=None,
):
    """ Marshal a list of Python objects to a CSV-compatible list of lists
        that can be passed to csv.writer.writerows

    Args:
        iterable: list, A list of objects that do not contain nested objects,
                  all fields must be of types in @types
        cls:      type-or-function,
        types:    tuple-of-types, The primitive types, typically
                  you would not change this
        fields:   None-or-list-of-str, Explicitly marshal only these fields
    Returns:
        dict
    """
    return [
        marshal_list(
            x,
            cls,
            types,
            fields,
        ) for x in iterable
    ]


def unmarshal_csv(
    iterable,
    cls,
):
    """ Unmarshal @iterable into a list of @cls

    Args:
        iterable: list-or-tuple-of-objects
        cls:      type-or-function, The type to unmarshal into
    Returns:
        List of instances of @cls
    Raises:
        ValueError: If @cls.__init__ does not contain a self argument
    """
    return [
        unmarshal_list(x, cls)
        for x in iterable
    ]



#!/usr/bin/env python3
"""

"""

import os
import setuptools
import sys

from setuptools.command.test import test as TestCommand


NAME = "pymarshal"
URL = 'https://github.com/j3ffhubb/pymarshal'
DESCRIPTION = (
    "Pythonic implementation of Golang's (un)marshalling of structs "
    "to/from various data serialization formats"
)


def _version():
    if 'test' in sys.argv:
        # avoid triggering a pytest coverage report bug
        return 'test'
    path = sys.path[:]
    dirname = os.path.dirname(__file__)
    abspath = os.path.abspath(dirname)
    sys.path.insert(
        0,
        abspath,
    )
    import pymarshal
    version = pymarshal.__version__
    sys.path = path
    return version

VERSION = _version()


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to pytest")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = [
            "--cov", NAME,
            "--cov-report", "html",
        ]

    def run_tests(self):
        import shlex
        #import here, because outside the eggs aren't loaded
        import pytest
        errno = pytest.main(shlex.split(self.pytest_args))
        sys.exit(errno)


setuptools.setup(
    name=NAME,
    version=VERSION,
    author="Pymarshal Team",
    author_email='j3ffhubb@users.noreply.github.com',
    license='BSD',
    description=DESCRIPTION,
    long_description=open('README.md', 'rt').read(),
    long_description_content_type='text/markdown',
    url=URL,
    packages=setuptools.find_packages(
        exclude=["*.test", "*.test.*", "test.*", "test"],
    ),
    include_package_data=True,
    install_requires=[],
    tests_require=[
        'bson',
        'pytest',
        'pytest-cov',
        'PyYAML',
    ],
    extras_require={
        'bson': ['bson']
    },
    cmdclass = {'test': PyTest},
    setup_requires=['pytest-runner'],
    # PyPI
    download_url="/".join([
        URL,
        "archive",
        "{}.tar.gz".format(VERSION),
    ]),
    keywords=[
        "go",
        "golang",
        "json",
        "bson",
        "yaml",
        "marshal",
        "unmarshal",
        "struct",
    ],
)



"""

"""

from .marshal import unmarshal_dict
from .key_swap import key_swap
import inspect


__all__ = [
    'type_assert',
    'type_assert_dict',
    'type_assert_iter',
]


def _msg(
    obj,
    cls,
):
    return '{0} is not an instance of {1}, is {2}'.format(
        obj,
        cls,
        type(obj),
    )


def _check_isinstance(obj, cls):
    if not isinstance(obj, cls):
        msg = _msg(obj, cls)
        raise TypeError(msg)


def _check(
    obj,
    cls,
    allow_none=False,
    cast_from=None,
    cast_to=None,
    dynamic=None,
    ctor=None,
    false_to_none=False,
    check=None,
):
    if (
        allow_none
        and
        obj is None
    ):
        return obj

    if (
        false_to_none
        and
        not obj
    ):
        return None

    if (
        obj is None
        and
        dynamic is not None
    ):
        _check_isinstance(dynamic, cls)
        obj = dynamic

    if (
        cast_from
        and
        isinstance(obj, cast_from)
    ):
        if cast_to:
            cast = cast_to(obj) if cast_to else cls(obj)
            _check_isinstance(cast, cls)
            obj = cast
        else:
            obj = cls(obj)

    if (
        check
        and
        not check(obj)
    ):
        try:
            source = inspect.getsource(check)
        except OSError:
            # When running in the Python terminal, getting the source code
            # will fail.  So just give the function and address.
            source = check
        msg = "'{}' failed check '{}'".format(obj, source)
        raise ValueError(msg)

    if not isinstance(obj, cls):
        if isinstance(obj, dict):
            obj = key_swap(obj, cls, True)
            new_obj = unmarshal_dict(obj, cls, ctor=ctor)
            _check_isinstance(new_obj, cls)
            obj = new_obj
        else:
            msg = _msg(obj, cls)
            raise TypeError(msg)

    return obj


def _check_dstruct(obj, cls):
    if cls is not None:
        _check_isinstance(obj, cls)


def _check_choices(obj, choices):
    if (
        choices is not None
        and
        obj not in choices
    ):
        msg = "{} not in {}".format(obj, choices)
        raise ValueError(msg)


def type_assert(
    obj,
    cls,
    allow_none=False,
    cast_from=None,
    cast_to=None,
    dynamic=None,
    choices=None,
    ctor=None,
    desc=None,
    false_to_none=False,
    check=None,
):
    """ Assert that @obj is an instance of @cls

        Note that you cannot use this to unmarshal JSON if @cls
        is a tuple of types, it can only be a single type.

        Will also unmarshal JSON objects to Python objects if @obj
        is an instance of dict and @cls is not dict

        Args:
            obj:        object instance, The object to type assert
            cls:        type,  The class type to assert
            allow_none: bool, True to allow '@obj is None', otherwise False
            cast_from:  type-or-tuple-of-types, If @obj is an instance
                        of this type(s), cast it to @cast_to
            cast_to:    type, The type to cast @obj to if it's an instance
                        of @cast_from, or None to cast to @cls.
                        If you need more than type(x), use a lambda or
                        factory function.
            dynamic:    @cls, A dynamic default value if @obj is None,
                        and @dynamic is not None.  @allow_none should be False
                        Valid uses:
                            datetime.datetime.now()
                            int(time.time())
                            # Or, to avoid the Python singleton bug when
                            # using arg=[], for example:
                            [], {}, set()
            choices:    iterable-or-None, If not None, @obj must
                        be in @choices
            ctor:       None-or-static-method, Use this method as the
                        constructor instead of __init__
            desc:       None-or-string, an optional description for this field,
                        for using this function to fully replace docstrings
            false_to_none: bool, True to cast falsey values such as "", 0, [],
                        to None
            check:      None-lambda-function, Single argument function to check
                        a value, return False if not valid, for example:
                            lambda x: x >= 0 and x < 256
        Returns:
            @obj
        Raises:
            TypeError:  If @obj is not an instance of @cls
            ValueError: If @check is not None and @obj fails @check
    """
    _check_choices(obj, choices)

    return _check(
        obj,
        cls,
        allow_none,
        cast_from,
        cast_to,
        dynamic=dynamic,
        ctor=ctor,
        false_to_none=false_to_none,
        check=check,
    )


def type_assert_iter(
    iterable,
    cls,
    cast_from=None,
    cast_to=None,
    dynamic=None,
    objcls=None,
    choices=None,
    ctor=None,
    allow_none=False,
    desc=None,
    false_to_none=False,
    check=None,
):
    """ Checks that every object in @iterable is an instance of @cls

        Will also unmarshal JSON objects to Python objects if items in
        @iterable are an instance of dict

        Args:
            iterable:   Any iterable to check.  Note that it would not
                        make sense to pass a generator to this function
            cls:        type, The class type to assert each member
                        of @iterable is
            cast_from:  type-or-tuple-of-types, If @obj is an instance
                        of this type(s), cast it to @cast_to
            cast_to:    type, The type to cast @obj to if it's an instance
                        of @cast_from, or None to cast to @cls.
                        If you need more than type(x), use a lambda or
                        factory function.
            dynamic:    @cls, A dynamic default value if @iterable is None,
                        and @dynamic is not None.
            objcls:     None-or-type, a type to assert @iterable is,
                        ie:  list, set, etc...
            choices:    iterable-or-None, If not None, each object in
                        @iterable must be in @choices
            ctor:       None-or-static-method: Use this method as the
                        constructor instead of __init__
            allow_none: bool, True to allow @iterable to be None,
                        otherwise False
            desc:       None-or-string, an optional description for this field,
                        for using this function to fully replace docstrings
            false_to_none: bool, True to cast falsey values such as "", 0, [],
                        to None
            check:      None-lambda-function, Single argument function to check
                        a value, return False if not valid, for example:
                            lambda x: x >= 0 and x < 256
        Returns:
            @iterable, note that @iterable will be recreated, which
            may be a performance concern if @iterable has many items
        Raises:
            TypeError:  If @obj is not an instance of @cls
            ValueError: If @check is not None and a value fails @check
    """
    if (
        allow_none
        and
        iterable is None
    ):
        return iterable

    _check_dstruct(iterable, objcls)

    if choices is not None:
        for obj in iterable:
            _check_choices(obj, choices)

    if (
        iterable is None
        and
        dynamic is not None
    ):
        iterable = dynamic
    t = type(iterable)
    return t(
        _check(
            obj,
            cls,
            False,
            cast_from,
            cast_to,
            ctor=ctor,
            false_to_none=false_to_none,
            check=check,
        ) for obj in iterable
    )


def type_assert_dict(
    d,
    kcls=None,
    vcls=None,
    allow_none=False,
    cast_from=None,
    cast_to=None,
    dynamic=None,
    objcls=None,
    ctor=None,
    desc=None,
    false_to_none=False,
    check=None,
):
    """ Checks that every key/value in @d is an instance of @kcls: @vcls

        Will also unmarshal JSON objects to Python objects if
        the value is an instance of dict and @vcls is a class type

        Args:
            d:          The dict to type assert
            kcls:       The class to type assert for keys.
                        NOTE: JSON only allows str keys
            vcls:       The class to type assert for values
            allow_none: Allow a None value for the values.
                        This would not make sense for the keys.
            cast_from:  type-or-tuple-of-types, If @obj is an instance
                        of this type(s), cast it to @cast_to
            cast_to:    type, The type to cast @obj to if it's an instance
                        of @cast_from, or None to cast to @cls.
                        If you need more than type(x), use a lambda or
                        factory function.
            dynamic:    @cls, A dynamic default value if @d is None,
                        and @dynamic is not None.
            objcls:     None-or-type, a type to assert @d is,
                        ie:  dict, etc...
                        Note that isinstance considers
                        collections.OrderedDict to be of type dict
            ctor:       None-or-static-method: Use this method as the
                        constructor instead of __init__
            desc:       None-or-string, an optional description for this field,
                        for using this function to fully replace docstrings
            false_to_none: bool, True to cast falsey values such as "", 0, [],
                        to None
            check:      None-lambda-function, Single argument function to check
                        a value, return False if not valid, for example:
                            lambda x: x >= 0 and x < 256
        Returns:
            @d, note that @d will be recreated, which
            may be a performance concern if @d has many items
        Raises:
            TypeError:  If a key is not an instance of @kcls or
                        a value is not an instance of @vcls
            ValueError: If @check is not None and a value fails @check
    """
    _check_dstruct(d, objcls)

    if (
        d is None
        and
        dynamic is not None
    ):
        d = dynamic
    t = type(d)
    return t(
        (
            _check(k, kcls) if kcls else k,
            _check(
                v,
                vcls,
                allow_none,
                cast_from,
                cast_to,
                ctor=ctor,
                false_to_none=false_to_none,
                check=check,
            ) if vcls else v,
        )
        for k, v in d.items()
    )




"""
    Replicates the functionality of Golang's
    struct (un)marshalling feature to/from JSON.

    For examples, see:
    https://github.com/j3ffhubb/pymarshal/blob/master/README.md
"""

from . import json
from .json import *

__version__ = '2.0.2'
__all__ = json.__all__
